@@ -58,14 +58,6 @@ enum {
 	CANCEL_CD_RECORD,
 };
 
-static GladeXML *xml;
-
-static NautilusBurnRecorder *recorder = NULL;
-static int                   cancel   = CANCEL_NONE;
-/* For the image spinning */
-static GList                *image_list;
-static GList                *files_for_cleanup = NULL;
-
 typedef struct {
 	char              *source_device;
 	char              *source_iso;
@@ -83,15 +75,24 @@ typedef struct {
 	gint64                         size;
 } NautilusBurnerWriteData;
 
+static GladeXML *xml;
+
+static NautilusBurnRecorder    *recorder = NULL;
+static int                      cancel   = CANCEL_NONE;
+/* For the image spinning */
+static GList                   *image_list;
+static GList                   *files_for_cleanup = NULL;
+static NautilusBurnerWriteData *write_data = NULL;
+
 static int write_disc (NautilusBurnerWriteData *data);
 
 static void
-nautilus_burner_write_data_free (NautilusBurnerWriteData *data)
+nautilus_burn_progress_write_data_free (NautilusBurnerWriteData *data)
 {
 	if (! data)
 		return;
 
-	nautilus_burn_drive_free (data->drive);
+	nautilus_burn_drive_unref (data->drive);
 
 	g_list_foreach (data->tracks, (GFunc)nautilus_burn_recorder_track_free, NULL);
 	g_list_free (data->tracks);
@@ -350,6 +351,45 @@ nautilus_burn_progress_temp_files_cleanup (void)
 	}
 }
 
+static void
+nautilus_burn_progress_set_write_data (NautilusBurnerWriteData *data)
+{
+	/* just set a global for now */
+	write_data = data;
+}
+
+static NautilusBurnerWriteData *
+nautilus_burn_progress_get_write_data (void)
+{
+	return write_data;
+}
+
+static void
+nautilus_burn_progress_cleanup (void)
+{
+	NautilusBurnerWriteData *data;
+
+	nautilus_burn_progress_image_cleanup ();
+	nautilus_burn_progress_temp_files_cleanup ();
+
+	data = nautilus_burn_progress_get_write_data ();
+	if (data)
+		nautilus_burn_progress_write_data_free (data);
+}
+
+static int
+nautilus_burn_progress_write_disc (void)
+{
+	NautilusBurnerWriteData *data;
+	int                      res;
+
+	data = nautilus_burn_progress_get_write_data ();
+
+	res = write_disc (data);
+
+	return res;
+}
+
 static void
 animation_changed_cb (NautilusBurnRecorder *recorder,
 		      gboolean              spinning,
@@ -415,49 +455,6 @@ progress_changed_cb (NautilusBurnRecorder *recorder,
 	nautilus_burn_progress_set_time (secs);
 }
 
-static char *
-add_desc (char       *string,
-	  const char *addition)
-{
-	char *new;
-
-	if (strcmp (string, "") == 0) {
-		new = g_strdup_printf ("%s", addition);
-	} else {
-		new = g_strdup_printf ("%s/%s", string, addition);
-	}
-	g_free (string);
-	return new;
-}
-
-static char *
-drive_type_get_supported_media (int type)
-{
-	char *string;
-
-	string = g_strdup ("");
-	if (type & NAUTILUS_BURN_DRIVE_TYPE_CD_RECORDER) {
-		string = add_desc (string, "CD-R");
-	}
-	if (type & NAUTILUS_BURN_DRIVE_TYPE_CDRW_RECORDER) {
-		string = add_desc (string, "CD-RW");
-	}
-	if (type & NAUTILUS_BURN_DRIVE_TYPE_DVD_RAM_RECORDER) {
-		string = add_desc (string, "DVD-RAM");
-	}
-	if (type & NAUTILUS_BURN_DRIVE_TYPE_DVD_RW_RECORDER) {
-		string = add_desc (string, "DVD-RW");
-	}
-	if (type & NAUTILUS_BURN_DRIVE_TYPE_DVD_PLUS_R_RECORDER) {
-		string = add_desc (string, "DVD+R");
-	}
-	if (type & NAUTILUS_BURN_DRIVE_TYPE_DVD_PLUS_RW_RECORDER) {
-		string = add_desc (string, "DVD+RW");
-	}
-
-	return string;
-}
-
 static gboolean
 drive_can_rewrite (const NautilusBurnDrive *drive)
 {
@@ -488,7 +485,7 @@ ask_for_media (const NautilusBurnDrive *drive,
 
 	mibs = (required_size + 1024 * 1024 - 1) / (1024 * 1024);
 
-	type_string = drive_type_get_supported_media (drive->type);
+	type_string = nautilus_burn_drive_get_supported_media_string (drive, TRUE);
 
 	can_rewrite = drive_can_rewrite (drive);
 
@@ -496,24 +493,24 @@ ask_for_media (const NautilusBurnDrive *drive,
 		msg = g_strdup (_("Please make sure another application is not using the drive."));
 		title = N_("Drive is busy");
 	} else if (!is_reload && can_rewrite) {
-		msg = g_strdup_printf (_("Please put a %s, with at least %d MiB free, into the drive."),
-				       type_string,
-				       mibs);
+		msg = g_strdup_printf (_("Please put a disc, with at least %d MiB free, into the drive.  The following disc types are supported:\n%s"),
+				       mibs,
+				       type_string);
 		title = N_("Insert a rewritable or blank disc");
 	} else if (!is_reload && !can_rewrite) {
-		msg = g_strdup_printf (_("Please put a %s, with at least %d MiB free, into the drive."),
-				       type_string,
-				       mibs);
+		msg = g_strdup_printf (_("Please put a disc, with at least %d MiB free, into the drive.  The following disc types are supported:\n%s"),
+				       mibs,
+				       type_string);
 		title = N_("Insert a blank disc");
 	} else if (can_rewrite) {
-		msg = g_strdup_printf (_("Please replace the disc in the drive with a %s, with at least %d MiB free."),
-				       type_string,
-				       mibs);
+		msg = g_strdup_printf (_("Please replace the disc in the drive with a supported disc with at least %d MiB free.  The following disc types are supported:\n%s"),
+				       mibs,
+				       type_string);
 		title = N_("Reload a rewritable or blank disc");
 	} else {
-		msg = g_strdup_printf (_("Please replace the disc in the drive with a %s, with at least %d MiB free."),
-				       type_string,
-				       mibs);
+		msg = g_strdup_printf (_("Please replace the disc in the drive with a supported disc with at least %d MiB free.  The following disc types are supported:\n%s"),
+				       mibs,
+				       type_string);
 		title = N_("Reload a blank disc");
 	}
 
@@ -805,44 +802,49 @@ setup_cancel_button (void)
 }
 
 static void
-write_another_disc_cb (GtkWidget               *widget,
-		       NautilusBurnerWriteData *data)
+write_another_disc_cb (GtkWidget *widget,
+		       gpointer   user_data)
 {
-	int res;
+	int                      res;
+	NautilusBurnerWriteData *data;
 
 	setup_cancel_button ();
 
+	data = nautilus_burn_progress_get_write_data ();
 	nautilus_burn_drive_eject ((NautilusBurnDrive *)data->drive);
 
-	res = write_disc (data);
+	res = nautilus_burn_progress_write_disc ();
 
 	if (res == NAUTILUS_BURN_RECORDER_RESULT_CANCEL) {
 		gtk_main_quit ();
 	}
-
-	nautilus_burner_write_data_free (data);
 }
 
 static void
-setup_close_button (NautilusBurnerWriteData *data)
+setup_close_button (void)
 {
-	GtkWidget *close_button;
-	GtkWidget *cancel_button;
+	GtkWidget               *close_button;
+	GtkWidget               *cancel_button;
+	NautilusBurnerWriteData *data;
 
 	cancel_button = glade_xml_get_widget (xml, "cancel_button");
 	close_button = glade_xml_get_widget (xml, "close_button");
 
 	gtk_widget_hide (cancel_button);
 	gtk_widget_show (close_button);
+	gtk_widget_grab_focus (close_button);
 
+	g_signal_handlers_disconnect_by_func (close_button, gtk_main_quit, NULL);
 	g_signal_connect (close_button, "clicked", G_CALLBACK (gtk_main_quit), NULL);
 
+	data = nautilus_burn_progress_get_write_data ();
 	if (data) {
 		GtkWidget *write_another_button;
 		write_another_button = glade_xml_get_widget (xml, "write_another_button");
 		gtk_widget_show (write_another_button);
 
-		g_signal_connect (write_another_button, "clicked", G_CALLBACK (write_another_disc_cb), data);
+		g_signal_handlers_disconnect_by_func (write_another_button, write_another_disc_cb, NULL);
+		g_signal_connect (write_another_button, "clicked", G_CALLBACK (write_another_disc_cb), NULL);
 	}
 }
 
@@ -884,7 +886,7 @@ select_iso_filename (void)
 	char      *filename;
 	int        response;
 
-	chooser = gtk_file_chooser_dialog_new_with_backend (_("Choose a filename for the cdrom image"),
+	chooser = gtk_file_chooser_dialog_new_with_backend (_("Choose a filename for the disc image"),
 							    NULL,
 							    GTK_FILE_CHOOSER_ACTION_SAVE,
 							    "unix",
@@ -900,7 +902,7 @@ select_iso_filename (void)
 	if (response == GTK_RESPONSE_OK) {
 		filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (chooser));
 		if (g_file_test (filename, G_FILE_TEST_EXISTS)) {
-			if (!overwrite_existing_file (GTK_WINDOW (chooser), filename)) {
+			if (! overwrite_existing_file (GTK_WINDOW (chooser), filename)) {
 				g_free (filename);
 				goto reselect;
 			}
@@ -1143,12 +1145,12 @@ write_disc (NautilusBurnerWriteData *data)
 	if (res == NAUTILUS_BURN_RECORDER_RESULT_FINISHED) {
 		nautilus_burn_progress_set_text (_("Complete"));
 
-		setup_close_button (data);
+		setup_close_button ();
 
 	} else if (res != NAUTILUS_BURN_RECORDER_RESULT_CANCEL) {
 		nautilus_burn_progress_set_text (_("An error occurred while writing"));
 
-		setup_close_button (data);
+		setup_close_button ();
 
 		show_error_message (recorder, error);
 		g_error_free (error);
@@ -1159,19 +1161,143 @@ write_disc (NautilusBurnerWriteData *data)
 	return res;
 }
 
+static void
+update_progress_dialog_disc_type (NautilusBurnDrive        *drive,
+				  NautilusBurnerSourceData *burn_data,
+				  NautilusBurnMediaType     type)
+{
+	GtkWidget *label;
+	GtkWidget *dialog;
+	GtkWidget *heading;
+	const char *label_text;
+	const char *title_text;
+	const char *heading_text;
+	char      *heading_markup;
+	int        input_type;
+	int        output_type;
+	/* Arrays structured as [input-type][output-type]
+	   input-type: files/default, image, drive
+	   output-type: disc/default, file, CD, DVD
+	 */
+	static const char *titles [][4] = {
+		{ N_("Writing Files to Disc"),
+		  N_("Writing Files to a Disc Image"),
+		  N_("Writing Files to CD"),
+		  N_("Writing Files to DVD")
+		},
+
+		{ N_("Writing Image to Disc"),
+		  N_("Copying Disc Image"),
+		  N_("Writing Image to CD"),
+		  N_("Writing Image to DVD")
+		},
+
+		{ N_("Copying Disc"),
+		  N_("Copying Disc to a Disc Image"),
+		  N_("Copying Disc to CD"),
+		  N_("Copying Disc to DVD")
+		}
+	};
+	static const char *headings [][4] = {
+		{ N_("Writing files to disc"),
+		  N_("Writing files to a disc image"),
+		  N_("Writing files to CD"),
+		  N_("Writing files to DVD")
+		},
+
+		{ N_("Writing image to disc"),
+		  N_("Copying disc image"),
+		  N_("Writing image to CD"),
+		  N_("Writing image to DVD")
+		},
+
+		{ N_("Copying disc"),
+		  N_("Copying disc to a disc image"),
+		  N_("Copying disc to CD"),
+		  N_("Copying disc to DVD")
+		}
+	};
+	static const char *descriptions [][4] = {
+		{ N_("The selected files are being written to a CD or DVD.  This operation may take a long time, depending on data size and write speed."),
+		  N_("The selected files are being written to a disc image file."),
+		  N_("The selected files are being written to a CD.  This operation may take a long time, depending on data size and write speed."),
+		  N_("The selected files are being written to a DVD.  This operation may take a long time, depending on data size and write speed.")
+		},
+
+		{ N_("The selected disc image is being written to a CD or DVD.  This operation may take a long time, depending on data size and write speed."),
+		  N_("The selected disc image is being copied to a disc image file."),
+		  N_("The selected disc image is being written to a CD.  This operation may take a long time, depending on data size and write speed."),
+		  N_("The selected disc image is being written to a DVD.  This operation may take a long time, depending on data size and write speed.")
+		},
+
+		{ N_("The selected disc is being copied to a CD or DVD.  This operation may take a long time, depending on data size and drive speed."),
+		  N_("The selected disc is being copied to a disc image file.  This operation may take a long time, depending on data size and drive speed."),
+		  N_("The selected disc is being copied to a CD.  This operation may take a long time, depending on data size and drive speed."),
+		  N_("The selected disc is being copied to a DVD.  This operation may take a long time, depending on data size and drive speed.")
+		}
+	};
+
+	if (! burn_data) {
+		input_type = 0;
+	} else if (burn_data->source_iso || burn_data->source_cue) {
+		input_type = 1;
+	} else if (burn_data->source_drive) {
+		input_type = 2;
+	} else {
+		input_type = 0;
+	}
+
+	if (drive->type == NAUTILUS_BURN_DRIVE_TYPE_FILE) {
+		output_type = 1;
+	} else {
+		switch (type) {
+		case NAUTILUS_BURN_MEDIA_TYPE_CD:
+		case NAUTILUS_BURN_MEDIA_TYPE_CDR:
+		case NAUTILUS_BURN_MEDIA_TYPE_CDRW:
+			output_type = 2;
+			break;
+		case NAUTILUS_BURN_MEDIA_TYPE_DVD:
+		case NAUTILUS_BURN_MEDIA_TYPE_DVDR:
+		case NAUTILUS_BURN_MEDIA_TYPE_DVDRW:
+		case NAUTILUS_BURN_MEDIA_TYPE_DVD_RAM:
+		case NAUTILUS_BURN_MEDIA_TYPE_DVD_PLUS_R:
+		case NAUTILUS_BURN_MEDIA_TYPE_DVD_PLUS_RW:
+		case NAUTILUS_BURN_MEDIA_TYPE_DVD_PLUS_R_DL:
+			output_type = 3;
+			break;
+		case NAUTILUS_BURN_MEDIA_TYPE_UNKNOWN:
+		default:
+			output_type = 0;
+			break;
+		}
+	}
+
+	label_text = descriptions [input_type][output_type];
+	title_text = titles [input_type][output_type];
+	heading_text = headings [input_type][output_type];
+
+	label   = glade_xml_get_widget (xml, "progress_description_label");
+	heading = glade_xml_get_widget (xml, "info_title_label");
+	dialog  = glade_xml_get_widget (xml, "progress_window");
+
+	gtk_label_set_text (GTK_LABEL (label), label_text);
+	gtk_window_set_title (GTK_WINDOW (dialog), title_text);
+
+	heading_markup = g_strdup_printf ("<big><b>%s</b></big>", heading_text);
+	gtk_label_set_markup (GTK_LABEL (heading), heading_markup);
+	g_free (heading_markup);
+}
+
 static int
-burn_cd (const NautilusBurnDrive *drive,
-	 const NautilusBurnDrive *source_drive,
-	 const char              *source_iso,
-	 const char              *source_cue,
-	 const char              *label,
-	 gint64                   burn_size,
-	 gboolean                 eject_cd,
-	 int                      speed,
-	 gboolean                 dummy,
-	 gboolean                 debug,
-	 gboolean                 overburn,
-	 gboolean                 burnproof)
+burn_cd (const NautilusBurnDrive  *drive,
+	 NautilusBurnerSourceData *burn_data,
+	 const char               *label,
+	 gboolean                  eject_cd,
+	 int                       speed,
+	 gboolean                  dummy,
+	 gboolean                  debug,
+	 gboolean                  overburn,
+	 gboolean                  burnproof)
 {
 	GError                        *error        = NULL;
 	char                          *filename     = NULL;
@@ -1181,6 +1307,16 @@ burn_cd (const NautilusBurnDrive *drive,
 	int                            fd;
 	int                            res      = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
 	gboolean                       try_overburn = FALSE;
+	const NautilusBurnDrive       *source_drive;
+	const char                    *source_iso;
+	const char                    *source_cue;
+	gint64                         burn_size;
+	NautilusBurnMediaType          media_type = NAUTILUS_BURN_MEDIA_TYPE_UNKNOWN;
+
+	source_drive = burn_data->source_drive;
+	source_iso   = burn_data->source_iso;
+	source_cue   = burn_data->source_cue;
+	burn_size    = burn_data->size;
 
 	filename = NULL;
 
@@ -1199,7 +1335,13 @@ burn_cd (const NautilusBurnDrive *drive,
 			if (path != NULL && path[0] != '\0') {
 				filename = g_build_filename (path, "image.iso.XXXXXX", NULL);
 				fd = g_mkstemp (filename);
-				close (fd);
+				if (fd < 0) {
+					g_error ("Could not create temporary filename for ISO image");
+					res = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
+					goto out;
+				} else {
+					close (fd);
+				}
 			}
 
 			g_free (path);
@@ -1211,23 +1353,29 @@ burn_cd (const NautilusBurnDrive *drive,
 	if (drive->type != NAUTILUS_BURN_DRIVE_TYPE_FILE) {
 		/* if we are reading from a drive then don't request media yet */
 		if (! source_drive) {
-			gboolean is_locked;
-			gboolean got_media;
+			gboolean              is_locked;
+			gboolean              got_media;
 
 			is_locked = nautilus_burn_drive_lock ((NautilusBurnDrive *)drive, _("Burning CD"), NULL);
 			got_media = request_media ((NautilusBurnDrive *)drive, burn_size, overburn, &try_overburn);
+
+			if (got_media)
+				media_type = nautilus_burn_drive_get_media_type (drive);
+
 			if (is_locked) {
 				nautilus_burn_drive_unlock ((NautilusBurnDrive *)drive);
 			} 
 
-			if (!got_media) {
+			if (! got_media) {
 				res = NAUTILUS_BURN_RECORDER_RESULT_CANCEL;
 				goto out;
 			}
-
 		}
 	}
 
+	/* update the description now that we may know the media type */
+	update_progress_dialog_disc_type (drive, burn_data, media_type);
+
 	if (source_iso == NULL
 	    && source_cue == NULL) {
 		NautilusBurnImageCreateFlags image_flags = 0;
@@ -1264,7 +1412,13 @@ burn_cd (const NautilusBurnDrive *drive,
 
 			filename = g_build_filename (g_get_tmp_dir (), "image.iso.XXXXXX", NULL);
 			fd = g_mkstemp (filename);
-			close (fd);
+			if (fd < 0) {
+				g_error ("Could not create temporary filename for ISO image");
+				res = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
+				goto out;
+			} else {
+				close (fd);
+			}
 
 			image_flags &= ~NAUTILUS_BURN_IMAGE_CREATE_WARN_SPACE;
 
@@ -1285,7 +1439,13 @@ burn_cd (const NautilusBurnDrive *drive,
 
 				filename = g_build_filename (g_get_home_dir (), ".image.iso.XXXXXX", NULL);
 				fd = g_mkstemp (filename);
-				close (fd);
+				if (fd < 0) {
+					g_error ("Could not create temporary filename for ISO image");
+					res = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
+					goto out;
+				} else {
+					close (fd);
+				}
 
 				image_flags |= NAUTILUS_BURN_IMAGE_CREATE_WARN_SPACE;
 
@@ -1325,7 +1485,7 @@ burn_cd (const NautilusBurnDrive *drive,
 		nautilus_burn_progress_set_time (-1);
 		g_free (str);
 
-		setup_close_button (NULL);
+		setup_close_button ();
 
 		res = NAUTILUS_BURN_RECORDER_RESULT_FINISHED;
 		goto out;
@@ -1368,19 +1528,23 @@ burn_cd (const NautilusBurnDrive *drive,
 		if (1) {
 			record_flags |= NAUTILUS_BURN_RECORDER_WRITE_BURNPROOF;
 		}
+		if (1) {
+			record_flags |= NAUTILUS_BURN_RECORDER_WRITE_DISC_AT_ONCE;
+		}
 
 		if (source_drive
 		    && strcmp (source_drive->device, drive->device) == 0) {
 			nautilus_burn_drive_eject ((NautilusBurnDrive *)source_drive);
 		}
 
-		data->drive  = nautilus_burn_drive_copy ((NautilusBurnDrive *)drive);
+		data->drive  = nautilus_burn_drive_ref ((NautilusBurnDrive *)drive);
 		data->tracks = g_list_prepend (data->tracks, track);
 		data->flags  = record_flags;
 		data->speed  = speed;
 		data->size   = burn_size;
 
-		res = write_disc (data);
+		nautilus_burn_progress_set_write_data (data);
+		res = nautilus_burn_progress_write_disc ();
 
 		if (source_iso == NULL) {
 			nautilus_burn_progress_temp_files_add (filename);
@@ -1900,8 +2064,6 @@ selection_dialog_response (GtkWidget                *dialog,
 		gboolean    eject;
 		gboolean                 dummy, debug, overburn, burnproof;
 		const char              *label;
-		const char              *label_format;
-		char                    *label_text;
 		int                      res, speed;
 		GConfClient             *gconf_client;
 
@@ -1915,14 +2077,10 @@ selection_dialog_response (GtkWidget                *dialog,
 		eject_check        = glade_xml_get_widget (xml, "eject_check");
 		speed_combobox     = glade_xml_get_widget (xml, "speed_combobox");
 		dummy_check        = glade_xml_get_widget (xml, "dummy_check");
+		progress_window    = glade_xml_get_widget (xml, "progress_window");
 
-		label_format = gtk_label_get_text (GTK_LABEL (description_label));
-		/* FIXME: get type of media in drive */
-		label_text = g_strdup_printf (label_format, _("CD or DVD"));
-		gtk_label_set_text (GTK_LABEL (description_label), label_text);
-		g_free (label_text);
+		update_progress_dialog_disc_type (drive, burn_data, NAUTILUS_BURN_MEDIA_TYPE_UNKNOWN);
 
-		progress_window = glade_xml_get_widget (xml, "progress_window");
 		gtk_window_set_icon_name (GTK_WINDOW (progress_window), "gnome-dev-cdrom");
 		nautilus_burn_progress_image_setup ();
 		gtk_widget_show (progress_window);
@@ -1948,11 +2106,8 @@ selection_dialog_response (GtkWidget                *dialog,
 		}
 
 		res = burn_cd (drive,
-			       burn_data->source_drive,
-			       burn_data->source_iso,
-			       burn_data->source_cue,
+			       burn_data,
 			       label,
-			       burn_data->size,
 			       eject,
 			       speed,
 			       dummy,
@@ -1960,8 +2115,6 @@ selection_dialog_response (GtkWidget                *dialog,
 			       overburn,
 			       burnproof);
 
-		nautilus_burn_progress_image_cleanup ();
-
 		if (res == NAUTILUS_BURN_RECORDER_RESULT_CANCEL) {
 			gtk_main_quit ();
 		}
@@ -2151,7 +2304,7 @@ main (int argc, char *argv[])
 	g_free (burn_data->source_device);
 	g_free (burn_data);
 
-	nautilus_burn_progress_temp_files_cleanup ();
+	nautilus_burn_progress_cleanup ();
 
 	return 0;
 }
